from typing import Dict, List, Set, Tuple

TraceList = List[Tuple[Dict[str, float], float]]


def write_run_trace(file: str, run_trace: TraceList, num_steps: int = 0, num_runs_per_step: int = 0) -> None:
    if num_steps <= 0 or num_runs_per_step <= 0:
        write_raw_run_trace(file, run_trace)
        return

    variable_list: List[str] = create_variable_list(run_trace)
    split_trace_list: List[List[TraceList]] = split_trace(
        run_trace, num_steps, num_runs_per_step)
    try:
        with open(file, "w") as writer:
            writer.write("@T,trace\n")
            writer.write("@H,cost,round,step,{}\n".format(",".join(variable_list)))
            round: int = 1
            for round_list in split_trace_list:
                step: int = 1
                for step_list in round_list:
                    for param, cost in step_list:
                        record = ",{},{},{}".format(cost, round, step)
                        for name in variable_list:
                            if name in param:
                                record += ",{}".format(param[name])
                            else:
                                record += ","
                        writer.write(record + "\n")
                    step = step + 1
                round = round + 1
    except IOError:
        print("Failed to write to file {}".format(file))


def write_raw_run_trace(file: str, run_trace: TraceList) -> None:
    variable_list: List[str] = create_variable_list(run_trace)
    try:
        with open(file, "w") as writer:
            writer.write("@T,trace\n")
            writer.write("@H,cost,{}\n".format(",".join(variable_list)))
            for param, cost in run_trace:
                record = ",{}".format(cost)
                for name in variable_list:
                    if name in param:
                        record += ",{}".format(param[name])
                    else:
                        record += ","
                writer.write(record + "\n")
    except IOError:
        print("Failed to write to file {}".format(file))


def create_variable_list(run_trace: TraceList) -> List[str]:
    variable_set: Set[str] = set()
    for param, cost in run_trace:
        for name in param.keys():
            variable_set.add(name)
    return list(variable_set)


def split_trace(run_trace: TraceList, num_steps: int, num_runs_per_step: int) -> List[List[TraceList]]:
    step = 1
    i = 0
    split_trace_list: List[List[TraceList]] = list()
    step_list: List[TraceList]
    while i < len(run_trace):
        if step == 1:
            step_list = list()
            split_trace_list.append(step_list)

        endpoint = i + num_runs_per_step
        if endpoint > len(run_trace):
            endpoint = len(run_trace)

        step_list.append(run_trace[i:endpoint])
        i = endpoint
        step = step + 1
        if step > num_steps:
            step = 1

    return split_trace_list



from tkinter import StringVar as sv
from tkinter import IntVar as iv
from tkinter import BooleanVar as bv
from tkinter import DoubleVar as dv
import json

class OptionManager():
    
    def __init__(self):
        
        self.project_data = {"name": "", "path": ""}
        self.arguments = {"param": [],
                 "url": sv(),
                 "mode": sv(),
                 "files": {},
                 "calibration_parameters": [],
                 "figure_parameters": []}
        self.steps = []
        self.service_parameters = {}
            
    def clear(self):
        self.arguments['param'].clear()
        self.arguments['url'].set("")
        self.arguments['mode'].set("Optimization: MG-PSO")
        self.arguments['files'] = {}
        self.arguments['calibration_parameters'].clear()
        self.arguments['figure_parameters'].clear()
        self.steps = []
        self.service_parameters = {}
    
    def add_arguments(self, arguments):
        
        if ("url" in arguments):
            self.arguments["url"].set(arguments["url"])
            
        if ("mode" in arguments):
            self.arguments["mode"].set(arguments["mode"])
            
        if ("files" in arguments):
            for file in arguments["files"]:
                name = file["name"]
                value = file["value"]
                obj = {"name": sv(), "value": sv()}
                obj["name"].set(name)
                obj["value"].set(value)
                self.arguments["files"][name] = obj
        
        for param in arguments["param"]:
            name = param["name"]
            value = param["value"]
            obj = {"name": sv(), "value": sv()}
            obj["name"].set(name)
            obj["value"].set(value)
            self.arguments["param"].append(obj)
            
        for param in arguments["calibration_parameters"]:
            name = param["name"]
            value = param["value"]
            obj = {"name": sv(), "value": sv()}
            obj["name"].set(name)
            obj["value"].set(value)
            self.arguments["calibration_parameters"].append(obj)
            
    def add_steps(self, steps):
        for step in steps:
            obj = {"param": [], "objfunc": [], "name": sv(), "message": sv(), "open": False}
            obj["name"].set("Group " + str(len(self.steps) + 1))
            obj["message"].set("Wow")
            
            for param in step["param"]:
                param_obj = {
                    "name": sv(), 
                    "bounds": (sv(), sv()), 
                    "default_value": sv(), 
                    "optimal_value": sv(), 
                    "type": sv(),
                    "calibration_strategy": sv()
                }
                param_obj["name"].set(param["name"])
                if "bounds" in param:
                    param_obj["bounds"][0].set(param["bounds"][0])
                    param_obj["bounds"][1].set(param["bounds"][1]) 
                else: 
                    param_obj["bounds"][0].set(0)
                    param_obj["bounds"][1].set(1)
                    
                if "type" in param:
                    param_obj["type"].set(param["type"])
                else:
                    param_obj["type"].set("float")
                    
                if "default_value" in param:
                    param_obj["default_value"].set(param["default_value"])
                else:
                    param_obj["default_value"].set(1)
                    
                if "optimal_value" in param:
                    param_obj["optimal_value"].set(param["optimal_value"])
                else:
                    param_obj["optimal_value"].set(0)
                    
                if "calibration_strategy" in param:
                    param_obj["calibration_strategy"].set(param["calibration_strategy"])
                else:
                    param_obj["calibration_strategy"].set("none")
                    
                obj["param"].append(param_obj)
            
            for objfunc in step["objfunc"]:
                objfunc_obj = {"name": sv(), "of": sv(), "weight": sv(), "data": (sv(), sv())}
                objfunc_obj["name"].set(objfunc["name"])
                objfunc_obj["of"].set(objfunc["of"])
                
                if ("weight" in objfunc): 
                    objfunc_obj["weight"].set(objfunc["weight"])
                else:
                    objfunc_obj["weight"].set(1)
                    
                objfunc_obj["data"][0].set(objfunc["data"][0])
                objfunc_obj["data"][1].set(objfunc["data"][1]) 
                obj["objfunc"].append(objfunc_obj)
            
            self.steps.append(obj)
    
    def add_function(self, step_index):
        obj = {"name": sv(), "of": sv(), "weight": sv(), "data": (sv(), sv())}
        obj["name"].set("ns")
        obj["of"].set("ns")
        obj["weight"].set(1)
        obj["data"][0].set("")
        obj["data"][1].set("")
        self.steps[step_index]["objfunc"].append(obj)
        
    def remove_function(self, step_index, index):
        self.steps[step_index]["objfunc"].pop(index)
        
    def dupe_function(self, step_index, index):
        my_func = self.steps[step_index]["objfunc"][index]
        
        new_object = {"name": sv(), "of": sv(), "weight": sv(), "data": (sv(), sv())}
        new_object["name"].set(my_func["name"].get())
        new_object["of"].set(my_func["of"].get())
        new_object["weight"].set(my_func["weight"].get())
        new_object["data"][0].set(my_func["data"][0].get())
        new_object["data"][1].set(my_func["data"][1].get())
        
        self.steps[step_index]["objfunc"].append(new_object)
    
    def add_bound(self, step_index, 
                  name="name", 
                  min=0, 
                  max=1,
                  type="float",
                  default_value=1,
                  optimal_value=0,
                  calibration_strategy="none"):
        obj = {
            "name": sv(), 
            "bounds": (sv(), sv()), 
            "default_value": sv(), 
            "optimal_value": sv(),
            "type": sv(),
            "calibration_strategy": sv()
        }
        obj["name"].set(name)
        obj["type"].set(type)
        obj["default_value"].set(default_value)
        obj["optimal_value"].set(optimal_value)
        obj["calibration_strategy"].set(calibration_strategy)
        obj["bounds"][0].set(min)
        obj["bounds"][1].set(max)
        self.steps[step_index]["param"].append(obj)
        
    def remove_bound(self, step_index, index):
        self.steps[step_index]["param"].pop(index)
        
    def add_argument(self, key, value):
        obj = {"name": sv(), "value": sv()}
        obj["name"].set(key)
        obj["value"].set(value)
        self.arguments["param"].append(obj)
        
    def add_calibration_param(self, key, value):
        obj = {"name": sv(), "value": sv()}
        obj["name"].set(key)
        obj["value"].set(value)
        self.arguments["calibration_parameters"].append(obj)

    def add_figure_param(self, key, value):
        obj = {"name": sv(), "value": sv()}
        obj["name"].set(key)
        obj["value"].set(value)
        self.arguments["figure_parameters"].append(obj)
    
    def move_argument_up(self, index):
        if index > 0:
            self.arguments["param"][index], self.arguments["param"][index - 1] = self.arguments["param"][index - 1], self.arguments["param"][index]
            
    def move_argument_down(self, index):
        if index < len(self.arguments["param"]) - 1:
            self.arguments["param"][index], self.arguments["param"][index + 1] = self.arguments["param"][index + 1], self.arguments["param"][index]
            
    def move_step_up(self, index):
        if index > 0:
            self.steps[index], self.steps[index - 1] = self.steps[index - 1], self.steps[index]
            
    def move_step_down(self, index):
        if index < len(self.steps) - 1:
            self.steps[index], self.steps[index + 1] = self.steps[index + 1], self.steps[index]
            
    def toggle_step_open(self, index):
        self.steps[index]["open"] = not self.steps[index]["open"]
            
    def remove_argument(self, index):
        self.arguments["param"].pop(index)
        
    def remove_calibration_parameter(self, index):
        self.arguments["calibration_parameters"].pop(index)

    def remove_figure_parameter(self, index):
        self.arguments["figure_parameters"].pop(index)
        
    def remove_step(self, index):
        self.steps.pop(index)
            
    def get_project_data(self):
        return self.project_data
            
    def set_path(self, filename):
        file_name = filename.split("/")[-1].replace(".json", "")
        path = filename.replace(file_name + ".json", "")
        self.project_data["path"] = path
        self.project_data["name"] = file_name
            
    def get_arguments(self):
        return self.arguments
    
    def get_steps(self):
        return self.steps
            
            
    def get_all_as_json(self):
        obj = {"arguments": self.arguments, "steps": self.steps}
        return obj
    
    def set_service_parameters(self, service_parameters):
        self.service_parameters = service_parameters
    
    def get_service_parameters(self):
        return self.service_parameters

    def get_metrics(self):
        result = {}
        result['arguments'] = {}
        result['calibration_parameters'] = []
        result['service_parameters'] = {}
        result['project_data'] = self.project_data
        for key, value in self.arguments.items():
            if key == 'url' or key == 'mode':
                result['arguments'][key] = value.get()
            elif key == 'files':
                result['arguments'][key] = {}
                #for name, obj in value.items():
                #    result['arguments'][key].append({'name': obj['name'].get(), 'value': obj['value'].get()})
            elif key == 'param':
                result['arguments'][key] = []
                for obj in value:
                    result['arguments'][key].append({'name': obj['name'].get(), 'value': obj['value'].get()})
            elif key == "calibration_parameters":
                #result['calibration_parameters'][key] = []
                for obj in value:
                    result['calibration_parameters'].append({'name': obj['name'].get(), 'value': obj['value'].get()})
        result['steps'] = []
        for step in self.steps:
            step_result = {}
            #step_result['name'] = step['name'].get()
            #step_result['open'] = step['open']
            step_result['param'] = []
            for param in step['param']:
                # try converting the bounds to numbers
                #try:
                if param['type'].get() == 'float':
                    step_result['param'].append(
                        {
                            'name': param['name'].get(), 
                            'bounds': (float(param['bounds'][0].get()), 
                                       float(param['bounds'][1].get())),
                            'default_value': float(param['default_value'].get()),
                            'optimal_value': float(param['optimal_value'].get()),
                            'type': 'float',
                            'calibration_strategy': param['calibration_strategy'].get()
                        }
                    )
                elif param['type'].get() == 'list':
                    step_result['param'].append(
                        {
                            'name': param['name'].get(), 
                            'bounds': (float(param['bounds'][0].get()), 
                                       float(param['bounds'][1].get())),
                            'default_value': param['default_value'].get(),
                            'optimal_value': param['optimal_value'].get(),
                            'type': 'list',
                            'calibration_strategy': param['calibration_strategy'].get()
                        }
                    )
                #except ValueError:
                #    step_result['param'].append(
                #        {
                #            'name': param['name'].get(), 
                #            'bounds': (param['bounds'][0].get(), 
                #                       param['bounds'][1].get())
                #        }
                #    )
            step_result['objfunc'] = []
            for objfunc in step['objfunc']:
                step_result['objfunc'].append({'name': objfunc['name'].get(), 'of': objfunc['of'].get(), 'weight': float(objfunc['weight'].get()), 'data': (objfunc['data'][0].get(), objfunc['data'][1].get())})
            result['steps'].append(step_result)
        return result



import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import pandas as pd
import numpy as np
import os
from PIL import Image, ImageTk
import customtkinter
import traceback

def generate_graphs(HomePage):
    try:
        selected_graph = HomePage.graph_selector_value.get()
        info = HomePage.option_manager.get_project_data()
        folder = os.path.join(info['path'], info['name'])
        if not os.path.exists(folder):
            os.makedirs(folder)
            
        if (selected_graph == "Best Cost Stacked"):
            HomePage.selected_graph_name = "best_cost_stacked"
            best_cost_stacked(HomePage.running_config['steps'], HomePage.progress_data, HomePage.option_manager)
        elif (selected_graph == "Best Cost by Round"):
            HomePage.selected_graph_name = "best_cost_by_round"
            best_cost_by_round(HomePage.running_config['steps'], HomePage.progress_data, HomePage.option_manager)
        elif (selected_graph == "Iteration Table"):
            HomePage.selected_graph_name = "table"
            table(HomePage.running_config['steps'], HomePage.progress_data, HomePage.option_manager)
        elif (selected_graph == "Calibrated Parameters"):
            HomePage.selected_graph_name = "calibrated_params_by_round"
            calibrated_params_by_round(HomePage.running_config['steps'], HomePage.calibration_data, HomePage.option_manager)
        elif (selected_graph == "Custom CSV"):
            HomePage.selected_graph_name = "custom_csv"
            custom_csv(HomePage, HomePage.option_manager)
        elif (selected_graph == "Compare CSV"):
            HomePage.selected_graph_name = "compare_csv"
            compare_csv(HomePage, HomePage.option_manager)
        elif (selected_graph == "Sampling CSV"):
            HomePage.selected_graph_name = "sampling_csv"
            sampling_csv(HomePage, HomePage.option_manager)
        elif (selected_graph == "Matrix Editor"):
            HomePage.selected_graph_name = "matrix_editor"
            matrix_editor(HomePage, HomePage.option_manager)
            pass
            
        image_path = os.path.join(folder, HomePage.selected_graph_name + ".png")
        
        if not os.path.exists(image_path):
            image_path = os.path.join("./images", "up.png")
        
        HomePage.graph_image_obj = Image.open(image_path)
        HomePage.graph_image = customtkinter.CTkImage(HomePage.graph_image_obj, size=(HomePage.image_width * HomePage.image_scale, HomePage.image_height * HomePage.image_scale))
        HomePage.graph_label.configure(image=HomePage.graph_image)
    except Exception as e:
        print(f"An exception occurred in Graph Generator: {str(e)}")
        print(f"Exception type: {type(e).__name__}")
        print("Traceback:")
        traceback.print_exc()

def best_cost_stacked(config, data, option_manager):
    fig = go.Figure()
        
    total_steps = len(config)
    
    # Get unique values from the round_step column of the dataframe
    for round_step in data.keys():
        fig.add_trace(go.Scatter(x=data[round_step]["percent"], y=data[round_step]["best_cost"], name=round_step))

    fig.update_layout(
        title="",
        xaxis_title="Progress (% through Group)",
        yaxis_title="Best Cost",
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            range=[0, 0.6],
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        )
    )
    
    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "best_cost_stacked.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "best_cost_stacked.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "best_cost_stacked.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "best_cost_stacked.html"), "w") as f:
        f.write(html)

    
        return fig

def table(config, dataframe, option_manager):
    # Create a plotly table with the values in the dataframe
    fig = go.Figure(data=[go.Table(
        header=dict(
            values=list(dataframe.columns),
            font=dict(color='black'),
            align="left"
        ),
        cells=dict(
            values=[dataframe[k].tolist() for k in dataframe.columns],
            font=dict(color='black'),
            align = "left")
    )])
    
    fig.update_layout(
        title="",
        xaxis_title="Iteration",
        yaxis_title="Best Cost",
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            range=[0, 0.6],
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        )
    )

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "table.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "table.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "table.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "table.html"), "w") as f:
        f.write(html)
    
    return fig

def best_cost_by_round(config, data, option_manager):
    fig = go.Figure()
        
    total_steps = len(config)
    
    # Get unique values from the round_step column of the dataframe
    for round_step in data.keys():
    #for iteration in dataframe['round_step'].unique():
        # Get best_cost and completed rounds rows for this iteration

        round_index = int(round_step.split(" -")[0].replace("Round: ", "")) - 1

        #df = dataframe[dataframe['round_step'] == iteration]
        
        #step_index = ((iteration) % total_steps)
        #round_index = ((iteration) // total_steps)

        fig.add_trace(go.Scatter(x=np.array(data[round_step]['percent']) + (100 * round_index), y=data[round_step]['best_cost'], name=round_step))
        
        #fig.add_trace(go.Scatter(x=df['completed_rounds'] + (df['total_rounds'] * round_index), y=df['best_cost'], name='Group ' + str(step_index + 1)))
        
        
        xx = np.max(np.array(data[round_step]["percent"]) + (100 * round_index))
        fig.add_shape(
            type='line',
            x0=xx,
            y0=0,
            x1=xx,
            y1=1,
            yref='paper',
            line=dict(
                color='rgb(102, 102, 102)',
                width=2
            )
        )
        
        fig.add_annotation(
            x=xx + 0.5,
            y=1,
            yref='paper',
            text='Round ' + str(round_index + 1),
            showarrow=False,
            yshift=-10
        )

    fig.update_layout(
        title="",
        xaxis_title="Progress (%)",
        yaxis_title="Best Cost",
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            range=[0, 0.6],
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        )
    )
    
    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "best_cost_by_round.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "best_cost_by_round.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "best_cost_by_round.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "best_cost_by_round.html"), "w") as f:
        f.write(html)
    
    return fig

def calibrated_params_by_round(config, list_of_objs, option_manager):
    fig = go.Figure()
        
    total_steps = len(config)
    
    datalines = {"step": [], "round": []}
    step = 1
    round = 1
    for index, obj in enumerate(list_of_objs):
        if (obj == {}):
            continue
        for key in obj.keys():
            if key not in datalines:
                datalines[key] = []
            datalines[key].append(obj[key])
        datalines["step"].append(step)
        datalines['round'].append(round)
        step += 1
        if (step > total_steps):
            step = 1
            round += 1
    
    # Get unique values from the round_step column of the dataframe
    for key in datalines.keys():
        # Get best_cost and completed rounds rows for this iteration
        if key == 'step' or key == 'round':
            continue
        
        fig.add_trace(go.Scatter(x=datalines['round'], y=datalines[key], name=key))

    fig.update_layout(
        title="",
        xaxis_title="Round",
        yaxis_title="Particle Parameters",
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        )
    )

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "calibrated_params_by_round.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "calibrated_params_by_round.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "calibrated_params_by_round.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "calibrated_params_by_round.html"), "w") as f:
        f.write(html)

    return fig

def custom_csv(homepage, option_manager):
    fig = go.Figure()

    data = homepage.csv_data["data"]

    x = homepage.csv_x_selector.get()
    val = homepage.csv_y1_selector.get()
    val2 = homepage.csv_y2_selector.get()
    
    xx = None
    if x == "time":
        xx = pd.to_datetime(data["time"], format='%Y-%m-%d', errors='coerce')
    elif x == "date":
        xx = pd.to_datetime(data["date"], format='%d-%m-%Y', errors='coerce')
    else:
        xx = pd.to_numeric(data[x], errors="coerce")
        
    yy = pd.to_numeric(data[val], errors="coerce")
    
    yy_unit = "-"
    if "Unit" in homepage.csv_data["data_attributes"]:
        yy_unit = homepage.csv_data["data_attributes"]["Unit"][val]
    
    yy2 = pd.to_numeric(data[val2], errors="coerce")
    
    yy2_unit = "-"
    if "Unit" in homepage.csv_data["data_attributes"]:
        yy2_unit = homepage.csv_data["data_attributes"]["Unit"][val2]

    fig.add_trace(go.Scatter(x=xx, y=yy, name=val))
    fig.add_trace(go.Scatter(x=xx, y=yy2, name=val2, yaxis='y2'))

    fig.update_layout(
        title="",
        xaxis_title=x,
        yaxis_title=val,
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            title=val + " (" + str(yy_unit) + ")",
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        ),
        yaxis2=dict(
            title=val2 + " (" + str(yy2_unit) + ")",
            overlaying='y',
            side='right'
        )
    )

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "custom_csv.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "custom_csv.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "custom_csv.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "custom_csv.html"), "w") as f:
        f.write(html)

    return fig

def compare_csv(homepage, option_manager):
    fig = go.Figure()

    data = homepage.csv_data["data"]
    data2 = homepage.csv_data2["data"]

    x = homepage.csv_x_selector.get()
    val = homepage.csv_y1_selector.get()
    val2 = homepage.csv_y2_selector.get()
    
    xx = None
    if x == "time":
        xx = pd.to_datetime(data["time"], format='%Y-%m-%d', errors='coerce')
    elif x == "date":
        xx = pd.to_datetime(data["date"], format='%d-%m-%Y', errors='coerce')
    else:
        xx = pd.to_numeric(data[x], errors="coerce")
        
    yy = pd.to_numeric(data[val], errors="coerce")
    
    xx2 = None
    if x == "time":
        xx2 = pd.to_datetime(data2["time"], format='%Y-%m-%d', errors='coerce')
    elif x == "date":
        xx2 = pd.to_datetime(data2["date"], format='%d-%m-%Y', errors='coerce')
    else:
        xx2 = pd.to_numeric(data2[x], errors="coerce")
    
    yy_unit = "-"
    if "Unit" in homepage.csv_data["data_attributes"]:
        yy_unit = homepage.csv_data["data_attributes"]["Unit"][val]
    
    yy2 = pd.to_numeric(data[val2], errors="coerce")
    
    yy2_unit = "-"
    if "Unit" in homepage.csv_data["data_attributes"]:
        yy2_unit = homepage.csv_data["data_attributes"]["Unit"][val2]

    fig.add_trace(go.Scatter(x=xx, y=yy, name=val))
    fig.add_trace(go.Scatter(x=xx2, y=yy2, name=val2, yaxis='y2'))

    fig.update_layout(
        title="",
        xaxis_title=x,
        yaxis_title=val,
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            title=val + " (" + str(yy_unit) + ")",
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        ),
        yaxis2=dict(
            title=val2 + " (" + str(yy2_unit) + ")",
            overlaying='y',
            side='right'
        )
    )

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "compare_csv.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "compare_csv.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "compare_csv.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "compare_csv.html"), "w") as f:
        f.write(html)

    return fig

def sampling_csv(homepage, option_manager):
    fig = go.Figure()

    style = homepage.figure_style.get()

    data = homepage.csv_data

    x = homepage.csv_x_selector.get()
    val = homepage.csv_y1_selector.get()
    val2 = homepage.csv_y2_selector.get()
    
    xx = None
    if x == "time":
        xx = pd.to_datetime(data["time"], format='%Y-%m-%d', errors='coerce')
    elif x == "date":
        xx = pd.to_datetime(data["date"], format='%d-%m-%Y', errors='coerce')
    else:
        xx = pd.to_numeric(data[x], errors="coerce")
        
    yy = pd.to_numeric(data[val], errors="coerce")
    
    yy_unit = ""
    
    yy2 = pd.to_numeric(data[val2], errors="coerce")
    
    yy2_unit = ""

    if (style == "Scatter"):
        fig.add_trace(go.Scatter(x=xx, y=yy, name=val, mode='markers'))
        fig.add_trace(go.Scatter(x=xx, y=yy2, name=val2, yaxis='y2', mode='markers'))
    elif (style == "Bars"):
        fig.add_trace(go.Bar(x=xx, y=yy, name=val))
        fig.add_trace(go.Bar(x=xx, y=yy2, name=val2, yaxis='y2'))
    elif (style == "Lines"):
        fig.add_trace(go.Scatter(x=xx, y=yy, name=val))
        fig.add_trace(go.Scatter(x=xx, y=yy2, name=val2, yaxis='y2'))
    elif (style == "Area"):
        fig.add_trace(go.Scatter(x=xx, y=yy, name=val, fill='tozeroy'))
        fig.add_trace(go.Scatter(x=xx, y=yy2, name=val2, yaxis='y2', fill='tozeroy'))
    elif (style == "Box"):
        fig.add_trace(go.Box(x=xx, y=yy, name=val))
        fig.add_trace(go.Box(x=xx, y=yy2, name=val2, yaxis='y2'))

    fig.update_layout(
        title="",
        xaxis_title=x,
        yaxis_title=val,
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)',
        xaxis=dict(
            gridcolor='rgb(72, 72, 72)',
            gridwidth=1
        ),
        yaxis=dict(
            title=val,
            autorange=True,
            gridcolor='rgb(72, 72, 72)',
            gridwidth=0.1
        ),
        yaxis2=dict(
            title=val2,
            overlaying='y',
            side='right'
        )
    )

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "sampling_csv.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "sampling_csv.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "sampling_csv.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "sampling_csv.html"), "w") as f:
        f.write(html)

    return fig

def matrix_editor(homepage, option_manager):
    style = homepage.figure_style.get()
    data = homepage.csv_data
    x = homepage.csv_x_selector.get()

    all_figures = []
    figure_parameters = option_manager.get_arguments()['figure_parameters']

    color_list = ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52']
    color_index = 0
    color_dict = {}

    total_figures = 0
    for parameter in figure_parameters:
        name = parameter['name'].get()
        if ("Fig" not in name):
            continue
        total_figures += 1

    figures_per_row = 1
    if (total_figures == 1):
        figures_per_row = 1
    elif (total_figures == 2):
        figures_per_row = 2
    elif (total_figures == 3 or total_figures == 4):
        figures_per_row = 2
    elif (total_figures == 5 or total_figures == 6):
        figures_per_row = 3
    elif (total_figures == 7 or total_figures == 8):
        figures_per_row = 4
    elif (total_figures == 9 or total_figures == 10):
        figures_per_row = 5

    fig_combined = make_subplots(rows=(total_figures // figures_per_row) + 1, cols=figures_per_row, shared_xaxes=False, shared_yaxes=False)

    for parameter in figure_parameters:
        name = parameter['name'].get()

        if ("Fig" not in name):
            continue

        fig = go.Figure()
        val = parameter['value'].get()
        
        xx = None
        if x == "time":
            xx = pd.to_datetime(data["time"], format='%Y-%m-%d', errors='coerce')
        elif x == "date":
            xx = pd.to_datetime(data["date"], format='%d-%m-%Y', errors='coerce')
        else:
            xx = pd.to_numeric(data[x], errors="coerce")
            
        yy = pd.to_numeric(data[val], errors="coerce")
        
        yy_unit = ""

        if (style == "Scatter"):
            fig.add_trace(go.Scatter(x=xx, y=yy, name=val, mode='markers'))
        elif (style == "Bars"):
            fig.add_trace(go.Bar(x=xx, y=yy, name=val))
        elif (style == "Lines"):
            fig.add_trace(go.Scatter(x=xx, y=yy, name=val))
        elif (style == "Area"):
            fig.add_trace(go.Scatter(x=xx, y=yy, name=val, fill='tozeroy'))
        elif (style == "Box"):
            fig.add_trace(go.Box(x=xx, y=yy, name=val))

        fig.update_layout(
            title="",
            xaxis_title=x,
            yaxis_title=val,
            font=dict(color='white'),
            paper_bgcolor='rgba(42, 42, 42, 0)',
            plot_bgcolor='rgb(62, 62, 62)',
            xaxis=dict(
                gridcolor='rgb(72, 72, 72)',
                gridwidth=1
            ),
            yaxis=dict(
                title=val,
                autorange=True,
                gridcolor='rgb(72, 72, 72)',
                gridwidth=0.1
            )
        )
        all_figures.append(fig)

    row = 1
    col = 1
    for fig in all_figures:
        for trace in fig.data:

            if trace.name not in color_dict:
                color_dict[trace.name] = color_list[color_index % len(color_list)]
                color_index += 1

            trace.marker.color = color_dict[trace.name]

            if fig_combined.data:
                trace.showlegend = not any(t.name == trace.name for t in fig_combined.data)
            fig_combined.add_trace(trace, row=row, col=col)
        
        # Get the layout titles from the individual figure
        xaxis_title = fig.layout.xaxis.title.text
        yaxis_title = fig.layout.yaxis.title.text

        # Update the combined figure's subplot with the titles
        fig_combined.update_xaxes(title_text=xaxis_title, row=row, col=col)
        fig_combined.update_yaxes(title_text=yaxis_title, row=row, col=col)

        col += 1
        if col > figures_per_row:
            row += 1
            col = 1

    info = option_manager.get_project_data()
    folder = os.path.join(info['path'], info['name'])
    
    fig = fig_combined
    fig.update_layout(
        title="",
        font=dict(color='white'),
        paper_bgcolor='rgba(42, 42, 42, 0)',
        plot_bgcolor='rgb(62, 62, 62)'
    )

    if not os.path.exists(folder):
        os.makedirs(folder)

    fig.write_image(os.path.join(folder, "matrix_editor.png"), width=1280, height=720)
    fig.write_html(os.path.join(folder, "matrix_editor.html"), include_plotlyjs='cdn', auto_open=False)
    with open(os.path.join(folder, "matrix_editor.html"), "r") as f:
        html = f.read()
        html = html.replace("<body>", "<body bgcolor='#2a2a2a'>")
    with open(os.path.join(folder, "matrix_editor.html"), "w") as f:
        f.write(html)

    return fig



import sys
from multiprocessing import Process, Queue
from queue import Empty
import threading
import time
import os
from .recosu.sampling.sampling import run_sampler
from .recosu.pso import global_best

def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()

def run_process(stdout_queue, stderr_queue, results_queue, data, folder):
    steps = data['steps']
    args = data['arguments']
    calib = data['calibration_parameters']
    
    my_mode = args["mode"]

    # If "mode" in args remove it
    if "mode" in args:
        del args["mode"]
    
    calibration_map = {}
    for param in calib:
        param_name = param['name']
        param_value = param['value']
        calibration_map[param_name] = param_value
        
    if not os.path.exists(folder):
        os.makedirs(folder)

    if not os.path.exists(os.path.join(folder, "results")):
        os.makedirs(os.path.join(folder, "results"))

    if (os.path.exists(os.path.join(folder, 'output.txt'))):
        os.remove(os.path.join(folder, 'output.txt'))
        
    if (os.path.exists(os.path.join(folder, 'error.txt'))):
        os.remove(os.path.join(folder, 'error.txt'))
    
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    
    read_stdout, write_stdout = os.pipe()
    read_stderr, write_stderr = os.pipe()
    
    sys.stdout = os.fdopen(write_stdout, 'w')
    sys.stderr = os.fdopen(write_stderr, 'w')
    
    stdout_thread = threading.Thread(target=enqueue_output, args=(os.fdopen(read_stdout, 'r'), stdout_queue))
    stderr_thread = threading.Thread(target=enqueue_output, args=(os.fdopen(read_stderr, 'r'), stderr_queue))
    stdout_thread.daemon = True
    stderr_thread.daemon = True
    stdout_thread.start()
    stderr_thread.start()
    
    try:
        
        options = {}
        oh_strategy = {}
        
        for key in calibration_map.keys():
            if "options_" in key:
                options[key.replace("options_", "")] = float(calibration_map[key])
            if "strategy_" in key:
                oh_strategy[key.replace("strategy_", "")] = calibration_map[key]

        config = {}

        if my_mode == "Sampling: Halton" or my_mode == "Sampling: Random":
            config = {
                'service_timeout': int(calibration_map['service_timeout']),
                'http_retry': int(calibration_map['http_retry']),
                'allow_redirects': True if calibration_map['allow_redirects'] == "True" else False,
                'async_call': True if calibration_map['async_call'] == "True" else False,
                'conn_timeout': int(calibration_map['conn_timeout']),
                'read_timeout': int(calibration_map['read_timeout']),
                'step_trace': os.path.join(folder, 'pso_step_trace.json')
            }
        else:
            config = {
                'service_timeout': int(calibration_map['service_timeout']),
                'http_retry': int(calibration_map['http_retry']),
                'http_allow_redirects': True if calibration_map['allow_redirects'] == "True" else False,
                'async_call': True if calibration_map['async_call'] == "True" else False,
                'http_conn_timeout': int(calibration_map['conn_timeout']),
                'http_read_timeout': int(calibration_map['read_timeout']),
                'particles_fail': int(calibration_map['particles_fail']),
                'step_trace': os.path.join(folder, 'pso_step_trace.json')
            }

        print("\n")
        print(steps)
        print("\n")
        print(args)
        print("\n")
        print(calibration_map)
        print("\n")
        print(options)
        print("\n")
        print(oh_strategy)
        print("\n")
        print(config)
        print("\n", flush=True)

        if my_mode == "Sampling: Halton":
            print("Running Halton Sampling..\n", flush=True)
            trace = run_sampler(steps, 
                                args, 
                                int(calibration_map['count']), 
                                int(calibration_map['num_threads']), 
                                "halton", 
                                conf=config, 
                                trace_file=os.path.join(folder, 'results', 'halton_trace.csv'),
                                offset=int(calibration_map['offset']))
            results_queue.put(trace)
            print(trace, flush=True)
            
        elif my_mode == "Sampling: Random":
            print("Running Random Sampling...\n", flush=True)
            trace = run_sampler(steps, 
                    args, 
                    int(calibration_map['count']), 
                    int(calibration_map['num_threads']), 
                    "random", 
                    conf=config, 
                    trace_file=os.path.join(folder, 'results', 'random_trace.csv'))
            results_queue.put(trace)
            print(trace, flush=True)
        else:
            print("Running MG-PSO Optimization...\n", flush=True)
            optimizer, trace = global_best(steps,   
                    rounds=(int(calibration_map['min_rounds']), int(calibration_map['max_rounds'])),              
                    args=args,      
                    n_particles=int(calibration_map['n_particles']),      
                    iters=int(calibration_map['iters']),  
                    n_threads=int(calibration_map['n_threads']),      
                    options=options,
                    oh_strategy=oh_strategy, 
                    conf=config
                )
            
            results_queue.put(trace)
            print(trace, flush=True)
        
        print("Finishing up...", flush=True)
        time.sleep(5)
    except Exception as e:
        print("An exception occurred: ", flush=True)
        print(str(e))
        # Print stack trace
        import traceback
        traceback.print_exc()

        # Write all of this information to a crash file
        with open(os.path.join(folder, 'crash.txt'), 'w') as f:
            f.write(str(e))
            f.write("\n")
            traceback.print_exc(file=f)
    finally:
        stdout_thread.join()
        stderr_thread.join()
        
        sys.stdout = old_stdout
        sys.stderr = old_stderr


import threading


class SampleTraceWriter:
    trace_file: str
    parameter_indices: dict[int, str]
    objective_indices: dict[int, str]
    write_lock: threading.Lock

    def __init__(self, trace_file: str):
        assert(trace_file is not None and len(trace_file) > 0)
        self.trace_file = trace_file
        self.parameter_indices = {}
        self.objective_indices = {}
        self.write_lock = threading.Lock()

    def write_header(self, parameter_names: list[str], objective_names: list[str]) -> None:
        with self.write_lock:
            with open(self.trace_file, 'w') as writer:
                writer.write("id")
                self.parameter_indices = {}
                index: int = 0
                for name in parameter_names:
                    writer.write(",{}".format(name))
                    self.parameter_indices[index] = name
                    index = index + 1
                self.objective_indices = {}
                index = 0
                for name in objective_names:
                    writer.write(",{}".format(name))
                    self.objective_indices[index] = name
                    index = index + 1
                writer.write("\n")

    def append_sample(self, sample_id: int, parameters: dict[str, any], objectives: dict[str, any]) -> None:
        with self.write_lock:
            with open(self.trace_file, 'a') as writer:
                writer.write("{}".format(sample_id))
                index: int = 0
                while index < len(self.parameter_indices):
                    writer.write(",{}".format(parameters[self.parameter_indices[index]]))
                    index = index + 1
                index = 0
                while index < len(self.objective_indices):
                    writer.write(",{}".format(objectives[self.objective_indices[index]]))
                    index = index + 1
                writer.write("\n")



import customtkinter
import json
import os
from multiprocessing import Process
import traceback
import re
import ast
import pandas as pd
import numpy as np

from ...util import PSORunner
from ...util import GraphGenerator
from ...util.CTkToolTip import CTkToolTip as ctt

def create_tab(self, tab):
    
    # URL
    tab.grid_columnconfigure(0, weight=1)
    tab.grid_rowconfigure(0, weight=1)
    tab.grid_rowconfigure(1, weight=200)

    #self.progress_container = customtkinter.CTkFrame(tab)
    #self.progress_container.grid_columnconfigure(0, weight=1)
    #self.progress_container.grid_columnconfigure(1, weight=1)
    #self.progress_container.grid_columnconfigure(2, weight=1)
    #self.progress_container.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    # Add progress bar to progress container
    #self.progress_message_left = customtkinter.CTkLabel(self.progress_container, text="")
    #self.progress_message_left.grid(row=0, column=0, padx=(10, 10), pady=(10, 10), sticky="w")
    
    #self.progress_message_middle = customtkinter.CTkLabel(self.progress_container, text="Calibration not running...")
    #self.progress_message_middle.grid(row=0, column=1, padx=(10, 10), pady=(10, 10), sticky="ew")
    
    #self.progress_message_right = customtkinter.CTkLabel(self.progress_container, text="")
    #self.progress_message_right.grid(row=0, column=2, padx=(10, 10), pady=(10, 10), sticky="e")
    
    self.textbox = customtkinter.CTkTextbox(tab)
    self.textbox.grid(row=1, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.textbox.insert("0.0", "Welcome to the CSIP PSO Calibration Tool!\n\nUse the Setup tab to define steps and calibration parameters. Use this tab to view logs and observe calibration progress. Once finished, use the Results tab to generate figures and graphs.")
        


import customtkinter
import tkinter as tk
from PIL import Image, ImageTk
import os
import platform
import subprocess

from . import SideBar

def create_tab(self, tab):
    
    def open_graph_in_browser():
        # Open the file in the default program
        info = self.option_manager.get_project_data()
        folder = os.path.join(info['path'], info['name'])
        
        if not os.path.exists(folder):
            os.makedirs(folder)
        
        file_path = os.path.join(folder, self.selected_graph_name + ".html")
        
        if platform.system() == "Windows":
            os.startfile(file_path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", file_path])
        else:
            subprocess.Popen(["xdg-open", file_path])
    
    def _resize_image(event):
        self.graph_label.update_idletasks()
        new_width = self.graph_label.winfo_width()
        new_height = self.graph_label.winfo_height()
        
        alt_width = new_height * 1.77778
        alt_height = new_width / 1.77778
        
        if (new_width < new_height):
            new_height = alt_height
        else:
            new_width = alt_width
        
        self.image_width = new_width
        self.image_height = new_height
        
        self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(new_width, new_height))
        self.graph_label.configure(image=self.graph_image)
        self.graph_label.update_idletasks()
    
    
    tab.grid_columnconfigure(0, weight=2)
    tab.grid_columnconfigure(1, weight=8)
    tab.grid_rowconfigure(0, weight=1)
    
    self.graph_sidebar = customtkinter.CTkFrame(tab)
    self.graph_sidebar.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    self.graph_container = customtkinter.CTkFrame(tab)
    self.graph_container.grid(row=0, column=1, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    self.graph_sidebar.grid_columnconfigure(0, weight=1)
    self.graph_sidebar.grid_rowconfigure(1, weight=1)
    self.graph_container.grid_columnconfigure(0, weight=1)
    self.graph_container.grid_rowconfigure(0, weight=1)
    
    self.graph_selector = customtkinter.CTkOptionMenu(self.graph_sidebar, values=["Best Cost Stacked", "Best Cost by Round", "Custom CSV", "Compare CSV", "Sampling CSV", "Matrix Editor"], variable=self.graph_selector_value, command=self.update_graph)
    self.graph_selector.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    # Create SideBar
    self.vis_sidebar = SideBar.SideBar(self.graph_sidebar, option_manager=self.option_manager, home_page=self, fg_color="transparent")
    self.vis_sidebar.grid(row=1, column=0, rowspan=8, padx=(0, 0), pady=(0, 0), sticky="nsew")
    
    # Add a button to call open_graph_in_browser
    self.graph_button = customtkinter.CTkButton(self.graph_sidebar, text="Open in Browser", command=open_graph_in_browser)
    self.graph_button.grid(row=9, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    
    self.graph_image_obj = Image.open(os.path.join("./images", "refresh_hd.png"))
    self.graph_image = customtkinter.CTkImage(self.graph_image_obj, size=(1280, 720))
    self.graph_label = customtkinter.CTkLabel(self.graph_container, text=None, image=self.graph_image)
    self.graph_label.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
    #window = customtkinter.CTk()
    #self.graph_label.bind('<Configure>', _resize_image)


from collections.abc import Iterable
import random


class RandomSampleGenerator:
    index: int
    count: int
    num_parameters: int
    rand: random.Random

    def __init__(self, count: int, num_parameters: int):
        assert (count > 0)
        self.index = 1
        self.count = count + 1
        self.num_parameters = num_parameters
        self.rand = random.Random()

    def __iter__(self) -> Iterable[tuple[int, list[float]]]:
        return self

    def __next__(self) -> tuple[int, list[float]]:
        if self.index >= self.count:
            raise StopIteration

        i: int = 0
        values: list[float] = []
        while i < self.num_parameters:
            values.append(self.rand.uniform(0, 1))
            i = i + 1

        result: tuple[int, list[float]] = (self.index, values)
        self.index = self.index + 1
        return result

